DEPRECATION_MSG_FUN = (
    "This function is deprecated and will be removed in a future release of "
    "`fractal-tasks-core`. Please migrate to `fractal-task-tools` for "
    "manifest building and task wrappers. See "
    "https://fractal-analytics-platform.github.io/fractal-task-tools/usage/"
    "#migrate-from-fractal-tasks-core."
)


DEPRECATION_MSG_CLASS = (
    "This class is deprecated and will be removed in a future release of "
    "`fractal-tasks-core`. Please migrate to `fractal-task-tools` for "
    "manifest building and task wrappers. See "
    "https://fractal-analytics-platform.github.io/fractal-task-tools/usage/"
    "#migrate-from-fractal-tasks-core."
)
